"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var ParserLib_1 = require("./ParserLib");
var Namespace_1 = require("../Namespace");
var ParserNamespace_1 = require("./ParserNamespace");
var TokenSpace_1 = require("../tokenizer/TokenSpace");
var TokenSet_1 = require("../tokenizer/TokenSet");
var Parser_1 = require("./Parser");
/** Parser configuration for quickly instantiating new parsers.
  * Each parser instance holds a new, cloned copy. */
var ParserConfig = /** @class */ (function () {
    /** XML parser configuration.
      * @param config Parent object for cloning.
      * @param native Reference to C++ object. For internal use only. */
    function ParserConfig(config, native) {
        // clonedNamespaceCount: number;
        this.registry = { tokens: {}, elements: {}, attributes: {} };
        if (config instanceof ParserConfig) {
            this.isLinked = true;
            this.options = config.options;
            this.uriSpace = config.uriSpace;
            this.prefixSpace = config.prefixSpace;
            this.elementSpace = config.elementSpace;
            this.attributeSpace = config.attributeSpace;
            this.xmlnsToken = config.xmlnsToken;
            this.emptyPrefixToken = config.emptyPrefixToken;
            this.xmlnsPrefixToken = config.xmlnsPrefixToken;
            this.processingPrefixToken = config.processingPrefixToken;
            this.uriSet = config.uriSet;
            this.prefixSet = config.prefixSet;
            this.namespaceList = config.namespaceList;
            this.namespaceTbl = config.namespaceTbl;
            this.maxNamespace = config.maxNamespace;
            this.nsMapper = config.nsMapper;
        }
        else {
            this.options = config || {};
            this.uriSpace = new TokenSpace_1.TokenSpace(7 /* uri */);
            this.prefixSpace = new TokenSpace_1.TokenSpace(8 /* prefix */);
            this.elementSpace = new TokenSpace_1.TokenSpace(9 /* element */);
            this.attributeSpace = new TokenSpace_1.TokenSpace(10 /* attribute */);
            this.xmlnsToken = this.attributeSpace.createToken('xmlns');
            this.uriSet = new TokenSet_1.TokenSet(this.uriSpace);
            this.prefixSet = new TokenSet_1.TokenSet(this.prefixSpace);
            this.namespaceList = [];
            this.namespaceTbl = {};
            this.maxNamespace = 0;
        }
        // this.clonedNamespaceCount = this.maxNamespace;
        if (!native) {
            this.emptyPrefixToken = this.prefixSet.createToken('');
            this.xmlnsPrefixToken = this.prefixSet.createToken('xmlns');
            this.processingPrefixToken = this.prefixSet.createToken('?');
            native = new ParserLib_1.NativeConfig(this.xmlnsToken.id, this.emptyPrefixToken.id, this.xmlnsPrefixToken.id, this.processingPrefixToken.id);
        }
        this.native = native;
        if (!this.isLinked && !this.options.omitDefaults) {
            this.bindNamespace(Namespace_1.Namespace.processing);
            this.bindNamespace(Namespace_1.Namespace.unknown);
            this.bindNamespace(Namespace_1.Namespace.xml1998);
        }
    }
    ParserConfig.prototype.unlink = function () {
        if (this.isFrozen)
            throw ('Cannot modify config object used by a parser.');
        if (!this.isLinked)
            return;
        this.isLinked = false;
        this.uriSpace = new TokenSpace_1.TokenSpace(7 /* uri */, this.uriSpace);
        this.prefixSpace = new TokenSpace_1.TokenSpace(8 /* prefix */, this.prefixSpace);
        this.elementSpace = new TokenSpace_1.TokenSpace(9 /* element */, this.elementSpace);
        this.attributeSpace = new TokenSpace_1.TokenSpace(10 /* attribute */, this.attributeSpace);
        this.uriSet = new TokenSet_1.TokenSet(this.uriSpace, this.uriSet);
        this.prefixSet = new TokenSet_1.TokenSet(this.prefixSpace, this.prefixSet);
        var namespaceList = this.namespaceList.slice(0);
        var num = namespaceList.length;
        while (num--) {
            var ns = namespaceList[num];
            // This just skips namespace 0 which never exists
            // (see ParserConfig.cc).
            if (ns) {
                ns = new ParserNamespace_1.ParserNamespace(ns, this);
                ns.id = num;
                namespaceList[num] = ns;
            }
        }
        var namespaceTbl = {};
        for (var _i = 0, _a = Object.keys(this.namespaceTbl); _i < _a.length; _i++) {
            var key = _a[_i];
            namespaceTbl[key] = namespaceList[this.namespaceTbl[key].id];
        }
        this.namespaceList = namespaceList;
        this.namespaceTbl = namespaceTbl;
    };
    ParserConfig.prototype.createParser = function () {
        var nativeParser = new ParserLib_1.NativeParser(this.native);
        var config = new ParserConfig(this, nativeParser.getConfig());
        this.isFrozen = true;
        return (new Parser_1.Parser(config, nativeParser));
    };
    ParserConfig.prototype.parseSync = function (data) {
        return (this.createParser().parseSync(data));
    };
    ParserConfig.prototype.getNamespace = function (uri) {
        var ns = this.namespaceTbl[uri];
        return (ns && ns.base);
    };
    ParserConfig.prototype.addNamespace = function (nsBase) {
        var uri = (this.nsMapper && this.nsMapper(nsBase.uri)) || nsBase.uri;
        var nsParser = this.namespaceTbl[uri];
        if (nsParser)
            return (nsParser.id);
        this.unlink();
        nsBase.uri = uri;
        nsParser = new ParserNamespace_1.ParserNamespace(nsBase, this);
        nsParser.id = this.native.addNamespace(nsParser.registerNative());
        this.namespaceList[nsParser.id] = nsParser;
        this.namespaceTbl[uri] = nsParser;
        if (nsBase.id > this.maxNamespace)
            this.maxNamespace = nsBase.id;
        if (nsBase.defaultPrefix)
            this.addPrefix(nsBase.defaultPrefix);
        this.addUri(uri, nsParser);
        return (nsParser.id);
    };
    ParserConfig.prototype.bindNamespace = function (nsBase, prefix, parser) {
        this.addNamespace(nsBase);
        var uri = (this.nsMapper && this.nsMapper(nsBase.uri)) || nsBase.uri;
        var nsParser = this.namespaceTbl[uri];
        if (!prefix && prefix != '')
            prefix = nsParser.base.defaultPrefix;
        (parser || this).bindPrefix(this.addPrefix(prefix), this.addUri(uri, nsParser));
        return (nsParser.id);
    };
    ParserConfig.prototype.updateNamespaces = function () {
        var list = this.namespaceList;
        var len = list.length;
        for (var num = 0; num < len; ++num) {
            if (list[num])
                list[num].registerNative();
        }
    };
    ParserConfig.prototype.bindPrefix = function (prefix, uri) {
        this.native.bindPrefix(prefix.id, uri.id);
    };
    ParserConfig.prototype.addUri = function (uri, ns) {
        this.unlink();
        var token = this.uriSet.createToken(uri);
        this.native.setUriTrie(this.uriSet.encodeTrie());
        this.native.addUri(token.id, ns.id);
        return (token);
    };
    ParserConfig.prototype.addPrefix = function (prefix) {
        this.unlink();
        var token = this.prefixSet.createToken(prefix);
        this.native.setPrefixTrie(this.prefixSet.encodeTrie());
        return (token);
    };
    ParserConfig.prototype.registerTokens = function (tbl) {
        var registry = this.registry;
        var token;
        var qname;
        for (var _i = 0, _a = Object.keys(tbl); _i < _a.length; _i++) {
            var prefix = _a[_i];
            var spec = tbl[prefix];
            var uri = spec.uri;
            var ns = this.getNamespace(uri) || new Namespace_1.Namespace(prefix, uri);
            for (var _b = 0, _c = spec.elements || []; _b < _c.length; _b++) {
                var name_1 = _c[_b];
                var tokens = this.getElementTokens(ns, name_1);
                token = tokens[0 /* open */];
                qname = prefix + ':' + name_1;
                registry.tokens[qname] = token;
                registry.elements[token.id] = qname;
            }
            for (var _d = 0, _e = spec.attributes || []; _d < _e.length; _d++) {
                var name_2 = _e[_d];
                var tokens = this.getAttributeTokens(ns, name_2);
                token = tokens[3 /* string */];
                qname = prefix + ':' + name_2;
                registry.tokens[qname] = token;
                registry.attributes[token.id] = qname;
            }
        }
        return (this.registry);
    };
    ParserConfig.prototype.jsxRegister = function (spec, handler) {
        var result = {};
        var token;
        for (var _i = 0, _a = Object.keys(spec); _i < _a.length; _i++) {
            var prefix = _a[_i];
            var elements = spec[prefix];
            var uri = elements.xmlns;
            var ns = this.getNamespace(uri) || new Namespace_1.Namespace(prefix, uri);
            result[prefix] = {};
            for (var _b = 0, _c = Object.keys(elements); _b < _c.length; _b++) {
                var name_3 = _c[_b];
                if (name_3 == 'xmlns')
                    continue;
                if (elements[name_3]) {
                    token = this.getElementTokens(ns, name_3)[0 /* open */];
                }
                else {
                    token = this.getAttributeTokens(ns, name_3)[3 /* string */];
                }
                result[prefix][name_3] = token;
            }
        }
        return (handler ? handler(result) : result);
    };
    ParserConfig.prototype.getElementTokens = function (ns, name) {
        var id = this.addNamespace(ns);
        return (this.namespaceList[id].addElement(name).tokenList);
    };
    ParserConfig.prototype.getAttributeTokens = function (ns, name) {
        var id = this.addNamespace(ns);
        return (this.namespaceList[id].addAttribute(name).tokenList);
    };
    return ParserConfig;
}());
exports.ParserConfig = ParserConfig;
