"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var Buffer_1 = require("../Buffer");
var InternalToken_1 = require("../parser/InternalToken");
var Node = /** @class */ (function () {
    function Node(token, buf, len, first, second) {
        this.token = token;
        this.buf = buf;
        this.len = len;
        this.first = first;
        this.second = second;
    }
    Node.prototype.clone = function () {
        var other = new Node(this.token, this.buf, this.len, this.first && this.first.clone(), this.second && this.second.clone());
        return (other);
    };
    return Node;
}());
/** Maximum number of bits per node (number must fit in 1 byte). */
var MAX_LEN = 255; // Test edge cases by using smaller numbers (>= 8) here!
/** Must equal Patricia :: notFound on C++ side. */
exports.NOT_FOUND = 0x7fffff;
var PatriciaCursor = /** @class */ (function () {
    function PatriciaCursor(node) {
        this.node = node;
        this.pos = 0;
        this.len = node.len;
    }
    PatriciaCursor.prototype.advance = function (c) {
        var node = this.node;
        var b = node.buf;
        var p = this.pos;
        var len = this.len;
        var delta = 0;
        while (len < 8) {
            if (len) {
                delta = (c ^ b[p++]) >> (7 - len);
            }
            else {
                if (!node.first)
                    return (false);
                delta = 0;
            }
            if (delta) {
                if (delta > 1) {
                    this.node = node;
                    this.pos = p - 1;
                    this.len = len;
                    return (false);
                }
                node = node.second;
            }
            else {
                node = node.first;
            }
            b = node.buf;
            p = 0;
            len = node.len;
        }
        if (c != b[p++]) {
            this.node = node;
            this.pos = p - 1;
            this.len = len;
            return (false);
        }
        len -= 8;
        this.node = node;
        this.pos = p;
        this.len = len;
        return (true);
    };
    return PatriciaCursor;
}());
var Patricia = /** @class */ (function () {
    function Patricia() {
    }
    Patricia.prototype.clone = function () {
        var other = new Patricia();
        if (this.root)
            other.root = this.root.clone();
        return (other);
    };
    Patricia.prototype.insertNode = function (token) {
        var pos = 0;
        var root = this.root;
        if (!token.name) {
            throw (new Error('Empty strings not supported'));
        }
        if (!root) {
            root = new Node(token, token.buf, token.buf.length * 8);
            this.root = root;
            return;
        }
        var cursor = new PatriciaCursor(root);
        while (pos < token.buf.length && cursor.advance(token.buf[pos]))
            ++pos;
        var node = cursor.node;
        var rest;
        if (pos < token.buf.length) {
            rest = new Node(token, token.buf.slice(pos), (token.buf.length - pos) * 8);
        }
        if (cursor.len) {
            var bit = 0;
            if (rest) {
                var c = token.buf[pos] ^ node.buf[cursor.pos];
                while (!(c & 0x80)) {
                    c <<= 1;
                    ++bit;
                }
            }
            else {
                // The new node is a prefix of this node.
                // Cut this node at a byte boundary.
            }
            // Split the node.
            node.first = new Node(node.token, node.buf.slice(cursor.pos), node.len - cursor.pos * 8, node.first, node.second);
            node.second = rest;
            node.token = rest ? null : token;
            node.buf = node.buf.slice(0, cursor.pos + ((bit + 7) >> 3));
            node.len = cursor.pos * 8 + bit;
        }
        else if (!rest) {
            throw (new Error('Duplicates not supported: ' + token.name));
        }
        else {
            // The new node only extends an existing node.
            node.first = rest;
        }
    };
    Patricia.prototype.insertList = function (tokenList) {
        for (var _i = 0, tokenList_1 = tokenList; _i < tokenList_1.length; _i++) {
            var token = tokenList_1[_i];
            this.insertNode(token);
        }
        // Verify that the tokens were correctly inserted!
        for (var _a = 0, tokenList_2 = tokenList; _a < tokenList_2.length; _a++) {
            var token = tokenList_2[_a];
            var pos = 0;
            var root = this.root;
            var cursor = new PatriciaCursor(root);
            while (pos < token.buf.length) {
                if (!cursor.advance(token.buf[pos++])) {
                    throw (new Error('Inserted token missing: ' + token.name));
                }
            }
            if (cursor.node.token != token) {
                throw (new Error('Wrong token inserted for: ' + token.name));
            }
        }
    };
    Patricia.encodeNode = function (node, dataList) {
        var len = node.len;
        var partLen;
        var byteLen;
        var totalByteLen = 0;
        var posIn = -1;
        var posOut;
        while (len) {
            partLen = len;
            if (partLen > MAX_LEN)
                partLen = MAX_LEN & ~7;
            // Convert bit to byte length rounding up, add 1 byte for length
            // header and 3 bytes for reference
            // (token ID or offset to second child).
            byteLen = (partLen + 7) >> 3;
            var data = new Buffer_1.ArrayType(byteLen + 4);
            dataList.push(data);
            totalByteLen += byteLen + 4;
            posOut = 0;
            data[posOut] = partLen;
            while (posOut < byteLen)
                data[++posOut] = node.buf[++posIn];
            var ref = void 0;
            if (len > MAX_LEN) {
                ref = exports.NOT_FOUND;
            }
            else {
                var nextTotalLen = 0;
                if (node.first)
                    nextTotalLen += Patricia.encodeNode(node.first, dataList);
                if (node.second) {
                    ref = nextTotalLen + 3;
                    nextTotalLen += Patricia.encodeNode(node.second, dataList);
                }
                else {
                    // ref = tokenSet.encode(node.token!) || 0;
                    ref = node.token.id;
                    if (!node.first)
                        ref |= 0x800000; // See 0x80 in PatriciaCursor.cc
                }
                totalByteLen += nextTotalLen;
            }
            data[++posOut] = ref >> 16;
            data[++posOut] = ref >> 8;
            data[++posOut] = ref;
            len -= partLen;
        }
        return (totalByteLen);
    };
    Patricia.prototype.encode = function () {
        var dataList = [];
        // Encode trie contents into a buffer.
        var dataLen = Patricia.encodeNode(this.root || Patricia.sentinel, dataList);
        return (Buffer_1.concatArray(dataList, dataLen));
    };
    /** Represents the root of an empty tree. */
    Patricia.sentinel = new Node(InternalToken_1.InternalToken.empty, InternalToken_1.InternalToken.empty.buf, InternalToken_1.InternalToken.empty.buf.length * 8);
    return Patricia;
}());
exports.Patricia = Patricia;
